#include <octave/oct.h>

DEFUN_DLD (evaluate_fast, args, ,
  "V=evaluate_fast(Xi,lambda,z,m,m0,CAT,Appendix,CENTER,H_mom,biH_mom,a_coeff,b_coeff)\n\
returns the value V of the thin-plate rbf at z, where z is a 1 x k complex vector interpreted\n\
as k points.  The returned vector V will be 1 x k and have error not exceeding epsilon.")
{
  octave_value_list retval;

  int nargin = args.length ();
  if (nargin != 12)
    {
      print_usage ();
      return retval;
    }/* if */

  ComplexRowVector Xi (args(0).complex_row_vector_value ());
  RowVector lambda ( args(1).vector_value ());
  ComplexRowVector z (args(2).complex_row_vector_value ());
  int m ( (int) args(3).double_value ()); 
  int m0 ( (int) args(4).double_value ()); 
  Matrix CATdb ( args(5).matrix_value ());       // This input is double precision, but changed to integer later
  Matrix Appendix ( args(6).matrix_value ());
  ComplexRowVector CENTER (args(7).complex_row_vector_value ());
  ComplexMatrix H_mom ( args(8).complex_matrix_value ());
  ComplexMatrix biH_mom ( args(9).complex_matrix_value ());
  Matrix a ( args(10).matrix_value ());             // Note: a instead of a_coeff
  Matrix b ( args(11).matrix_value ());             // Note: b instead of b_coeff

  Complex w, temp, temp1, z_wig;
  double s, s2, log_s;
  int N_z = z.length ();
  int N = CENTER.length ();
  RowVector V(N_z);
  int sp, stack[N], k, j, i, sum_flag;
  int j1, j2, l;
  int CAT[8][N];
  double t, v;
  for (j=0;j<N;j++){
    for (k=0;k<4; k++)
      CAT[k][j]=(int) CATdb(k,j)-1;
    CAT[4][j]=(int) CATdb(4,j);
    CAT[5][j]=(int) CATdb(5,j)-1;
    CAT[6][j]=(int) CATdb(6,j)-1;
    CAT[7][j]=(int) CATdb(7,j);
  } // for
//                      Here begins the outermost loop, where the evaluation points are processed one by one.
//                      In section 6 of the paper, the evaluation point is z and the obtained approximation to
//                      s(z) is V, here the evalution point is z(k) and the obtained value is v.
for (k=0; k<N_z; k++){
  sp=0;            // number of entries in the stack
  v=0;
  stack[sp++]=0;
//                                     Now the fun begins...  
//                      The stack has been initialized as holding only page 0 and v has been initialized to 0.
//                      For each iteration of the following while loop, a page number (called j)
//                      is popped off the stack and then visited.  As mentioned at the beginning of section 6,
//                      the visit to page j results in one of the following actions:
//                      (i) a value is added to v, or
//                      (ii) the children of page j are pushed onto the stack.
//                      Note that (i) can happen in one of two ways: 
//                             1. the summarized value of the cluster sum is accepted and added to v or
//                             2. the summarized value is not accepted, but j has no children, so the cluster sum
//                                is computed directly and added to v.
  while (sp>0){
    sum_flag=0;
    j= stack[--sp]; l=CAT[7][j];   // j has been popped off the stack and l holds the level of page j
//                      We first have to determine whether the evaluation point lies inside or outside the bounding disc.
//                      This is done by comparing the distance s from the evaluation point to the center of cluster j
//                      with the radius r_l of the bounding disc (note s2=s^2 and r_l^2 is stored in the Appendix).
//                      Note above that sum_flag has been initialized to 0.  This flag will get set to 1 if a summary is accepted.
    w=z(k)-CENTER(j); s2=real(w)*real(w) + imag(w)*imag(w);
    if (s2>=Appendix(l,1)) {   
//                      Case 1: z lies outside bounding disc
//                        Then the outer summary may or may not be acceptable.  This is decided by comparing s with the outer
//                        threshold radius T_ell (available in the Appendix)
      if (s2>=Appendix(l,2)){ 
//                      Case 1a: z lies outside bounding disc and outer summary is accepted.
//                        We compute the outer summary and add it to v.
        sum_flag=1;
        z_wig=w/Appendix(l,3);
        s2=s2/Appendix(l,1); // s2=real(z_wig)^2 + imag(z_wig)^2;
        log_s = 0.5*log(s2);
        temp1=1.0/z_wig;
        temp=temp1;
        t=real(H_mom(0,j)*s2*log_s+biH_mom(0,j)*(1+log_s));
        t+=real(temp*(-H_mom(1,j)*s2*(1+2*log_s)-0.5*biH_mom(1,j)));
        for (i=2; i<=m; i++){
          temp=temp*temp1;
          t+=real(temp*(H_mom(i,j)*s2-biH_mom(i,j)));
        } // for
        v+=Appendix(l,4)*real(H_mom(0,j)*s2-2.0*z_wig*conj(H_mom(1,j))+biH_mom(0,j))+Appendix(l,1)*t;
      } //              end of Case 1a.
    } //                end of Case 1.
    else if (s2>=Appendix(l,0)){
//                      Case 2: z lies inside bounding disc and the inner summary is accepted
//                         We compute the inner summary and add it to v.
        sum_flag=1;
        z_wig=w/Appendix(l,3);
        s2=s2/Appendix(l,1); // s2=real(z_wig)^2 + imag(z_wig)^2;
        s=sqrt(s2);
        if (s>0)
          temp1=conj(z_wig)/s;
        else
          temp1=1.0;
        temp=temp1;
        // w=polyval(a_coeff(:,k+1),s)*real(H_mom(1,p));
        i=0;
        t=(a(5,i)+s*(a(4,i)+s*(a(3,i)+s*(a(2,i)+s*(a(1,i)+s*a(0,i))))))*real(H_mom(0,j));
        t+=(b(5,i)+s*(b(4,i)+s*(b(3,i)+s*(b(2,i)+s*(b(1,i)+s*b(0,i))))))*real(biH_mom(0,j));
        i=1;
        t+=s*(a(5,i)+s*(a(4,i)+s*(a(3,i)+s*(a(2,i)+s*(a(1,i)+s*a(0,i))))))*real(temp*H_mom(1,j));
        t+=s*(b(5,i)+s*(b(4,i)+s*(b(3,i)+s*(b(2,i)+s*(b(1,i)+s*b(0,i))))))*real(temp*biH_mom(1,j));
        for (i=2; i<=m0; i++){
          temp*=temp1;
          t+=s2*(a(5,i)+s*(a(4,i)+s*(a(3,i)+s*(a(2,i)+s*(a(1,i)+s*a(0,i))))))*real(temp*H_mom(i,j));
          t+=s2*(b(5,i)+s*(b(4,i)+s*(b(3,i)+s*(b(2,i)+s*(b(1,i)+s*b(0,i))))))*real(temp*biH_mom(i,j));
        } // for
        v+=Appendix(l,4)*real(H_mom(0,j)*s2-2.0*z_wig*conj(H_mom(1,j))+biH_mom(0,j))+Appendix(l,1)*t;
      } //           end of Case 2
    if (sum_flag==0) {
//                   Case 3: The summary is not accepted.
      if (CAT[4][j]==0){ 
//                   Case 3a: Page j has no children
//                      We directly evaluate the cluster sum and add it to v.
        j1=CAT[5][j]; j2=CAT[6][j];
        for (i=j1; i<=j2; i++){
          //t=(x0-Xi(0,j))*(x0-Xi(0,j)) + (x1-Xi(1,j))*(x1-Xi(1,j));
          w=z(k)-Xi(i);
          t=real(w)*real(w) + imag(w)*imag(w);
          if (t>0)
            v+=0.5*t*log(t)*lambda(i);
        } // for j
      } //           end of Case 3a
      else{
//                   Case 3b: Page j has children
//                      We push the children of page j onto the stack; nothing is added to v.
        if (CAT[0][j]>=0)
          stack[sp++]=CAT[0][j];
        if (CAT[1][j]>=0)
          stack[sp++]=CAT[1][j];
        if (CAT[2][j]>=0)
          stack[sp++]=CAT[2][j];
        if (CAT[3][j]>=0)
          stack[sp++]=CAT[3][j];
      } //           End of Case 3b
    } //             End of Case 3
  } //               End of all the fun (while loop)
  V(k)=v;
}//                  End of the main loop (for loop)
retval(0)= V;

return retval;
}
